from typing import Dict, List, MutableMapping, Tuple, Union

from uuid import uuid4


# Only allow storing built-in types because other objects are tricky to
# (de)serialize when their structure changes.
DatabaseValue = Union[
    int, float, str,
    Tuple['DatabaseValue'],
    List['DatabaseValue'],
    Dict[str, 'DatabaseValue'],
]

DatabaseStore = MutableMapping[str, DatabaseValue]


class Database:
    def __init__(self, store: DatabaseStore):
        self._store = store

    @property
    def app_id(self) -> str:
        default = str(uuid4())
        identifier = self._store.setdefault('app_id', default)
        return str(identifier)

    @property
    def auth_token(self) -> str:
        return str(self._store['auth_token'])

    @auth_token.setter
    def auth_token(self, value: str):
        self._store['auth_token'] = value

    def content(self) -> DatabaseStore:
        return dict(self._store.items())



from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta
import logging
import threading
from typing import Dict, NamedTuple, Optional, Tuple

from plexapi.base import Playable
from plexapi.client import PlexClient
from plexapi.server import PlexServer
from plexapi.video import Episode
from typing_extensions import Literal
from wrapt import synchronized

from .notifications import NotificationContainerDict, PlaybackNotification


logger = logging.getLogger(__name__)

SessionKey = str


@dataclass(frozen=True, eq=False)
class Session:
    key: str
    state: Literal['buffering', 'playing', 'paused', 'stopped']
    playable: Playable
    player: PlexClient

    @classmethod
    def from_playable(cls, playable: Playable) -> 'Session':
        player = playable.players[0]
        return cls(
            key=str(playable.sessionKey),
            state=player.state,
            playable=playable,
            player=player,
        )

    def __hash__(self):
        return hash(self.key)

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.key == other.key


class IntroMarker(NamedTuple):
    # In milliseconds.
    start: int
    end: int


@dataclass(frozen=True, eq=False)
class EpisodeSession(Session):
    playable: Episode
    view_offset_ms: int

    @classmethod
    def from_playable(cls, episode: Episode) -> 'EpisodeSession':
        assert not episode.isFullObject()  # Probably dangerous wrt viewOffset otherwise.
        player = episode.players[0]

        return cls(
            key=str(episode.sessionKey),
            state=player.state,
            playable=episode,
            player=player,
            view_offset_ms=int(episode.viewOffset),
        )

    def intro_marker(self) -> Optional[IntroMarker]:
        if not self.playable.hasIntroMarker:
            return None
        internal = next(m for m in self.playable.markers if m.type == 'intro')
        return IntroMarker(start=internal.start, end=internal.end)


class SessionFactory:
    @classmethod
    def make(cls, playable: Playable) -> Session:
        if isinstance(playable, Episode):
            return EpisodeSession.from_playable(playable)
        return Session.from_playable(playable)


class SessionListener(ABC):
    def accept_session(self, session: Session) -> bool:
        """See the docstrings for the callback methods."""
        return True

    @abstractmethod
    def on_session_activity(self, session: Session):
        """Called iff accept_session(session) returns True."""
        pass

    @abstractmethod
    def on_session_removal(self, session: Session):
        """Called iff on_session_activity(session) was called at some point."""
        pass


class SessionExtrapolator(ABC):
    @abstractmethod
    def trigger_extrapolation(self, session: Session, listener_accepted: bool) -> bool:
        pass

    @abstractmethod
    def extrapolate(self, session: Session) -> Tuple[Session, int]:
        """
        Returns the extrapolated session and the extrapolation delay in ms.
        Called iff trigger_extrapolation(session) returns True.
        """
        pass


class SessionDispatcher:
    def __init__(self, listener: SessionListener, removal_timeout_sec: int = 20):
        # Plex's session WebSocket announces every 10 second, so we
        # conservatively set the default removal_timeout_sec to 20 seconds.
        self._listener = listener
        self._removal_timeout_sec = removal_timeout_sec
        # Sessions to track and potentially remove after a period of
        # removal_timeout_sec with no dispatching attempt.
        self._last_active: Dict[Session, datetime] = {}

    def dispatch(self, session: Session) -> bool:
        """
        Dispatches the session if listener.accept_session(session) is True.
        Returns the result of that call.
        """
        if isinstance(session, EpisodeSession) and session not in self._last_active:
            logger.info(
                f'New session {session.key}: {session.player} is playing {session.playable} '
                f'(intro marker = {session.playable.hasIntroMarker})'
            )

        accepted = False
        if self._listener.accept_session(session):
            accepted = True
            self._listener.on_session_activity(session)

        now = datetime.now()
        self._last_active[session] = now

        # Remove sessions that we haven't seen in the last removal_timeout_sec
        # period, in case dispatch_removal() wasn't called for some reason.
        timeout_ago = now - timedelta(seconds=self._removal_timeout_sec)
        for s, last_active in list(self._last_active.items()):
            if last_active <= timeout_ago:
                self._dispatch_removal(s)

        return accepted

    def _dispatch_removal(self, session: Session):
        if isinstance(session, EpisodeSession):
            logger.info(f'Session {session.key} ended: {session.player} stopped playing {session.playable}')
        self._listener.on_session_removal(session)
        del self._last_active[session]

    def dispatch_removal(self, removed_key: SessionKey) -> bool:
        for s in list(self._last_active.keys()):
            if s.key == removed_key:
                self._dispatch_removal(s)
                return True
        return False


class SessionNotFoundError(Exception):
    """Raised by SessionProvider when it could not find a session."""
    pass


class SessionProvider:
    def __init__(self, server: PlexServer):
        self._server = server

    def provide(self, session_key: str) -> Session:
        """Raises SessionNotFoundError when the session could not be found."""
        sessions = self._server.sessions()
        playable: Playable
        for playable in sessions:
            if str(playable.sessionKey) == session_key:
                return SessionFactory.make(playable)
        raise SessionNotFoundError(f'could not find session key {session_key} among {sessions}')


class SessionDiscovery:
    def __init__(
        self,
        server: PlexServer,
        provider: SessionProvider,
        dispatcher: SessionDispatcher,
        extrapolator: SessionExtrapolator,
    ):
        self._server = server
        self._provider = provider
        self._dispatcher = dispatcher
        self._extrapolator = extrapolator

        # To avoid leaks, preserve the following invariant:
        # timer in dict <=> timer alive,
        # where alive = started and not (done executing or cancelled).
        self._timers: Dict[SessionKey, threading.Timer] = {}

    @synchronized
    def alert_callback(self, alert: NotificationContainerDict):
        if alert['type'] == 'playing':
            # Never seen a case where the alert doesn't contain exactly one
            # notification, but let's loop over the list out of caution.
            for notification in alert['PlaySessionStateNotification']:  # type: ignore
                self._handle_notification(notification)

    @synchronized
    def _dispatch_and_schedule_extrapolated(self, session: Session):
        """Dispatches the specified session and potentially extrapolates it."""
        accepted = self._dispatcher.dispatch(session)

        # Preserve the timers invariant: this thread will die if this session
        # doesn't trigger an extrapolation. Note there won't be a dict entry
        # if this function wasn't called as part of a timer, so popping nothing
        # is fine.
        self._timers.pop(session.key, None)

        if not self._extrapolator.trigger_extrapolation(session, accepted):
            logger.debug(f'Will not extrapolate session {session}')
            return

        assert session.key not in self._timers
        new_session, delay_ms = self._extrapolator.extrapolate(session)
        delay_sec = delay_ms / 1000
        new_timer = threading.Timer(
            delay_sec,
            self._dispatch_and_schedule_extrapolated,
            args=(new_session,),
        )
        new_timer.daemon = True
        new_timer.start()
        self._timers[new_session.key] = new_timer

        logger.debug(
            f'Timer (delay={delay_sec:.3f}s) started for extrapolated session '
            f'{new_session} (original: {session})'
        )

    @synchronized
    def _handle_notification(self, notification: PlaybackNotification):
        # Dispatch regular notifications and simulate the rest while extrapoling
        # viewOffset using a timer. When a regular notification comes in, we
        # stop the active timer and handle the notification, then the process
        # repeats.

        # Ensure this is a string because I don't trust the Plex API.
        session_key = str(notification['sessionKey'])
        logger.debug(
            f'Incoming notification for session key {session_key} '
            f'(state = {notification["state"]})'
        )

        # Incoming regular notification, stop the active timer if any. And even
        # though we might recreate one on the spot, let's also remove the dict
        # entry to prevent any leak.
        old_timer = self._timers.pop(session_key, None)
        if old_timer:
            old_timer.cancel()
            logger.debug(f'Cancelled timer for session key {session_key}')
        else:
            logger.debug(f'No existing timer for session key {session_key}')

        if notification['state'] == 'stopped':
            # The HTTP API won't contain the session anymore, so just dispatch
            # the removal and return.
            self._dispatcher.dispatch_removal(session_key)
            return

        try:
            session = self._provider.provide(session_key)
        except SessionNotFoundError:
            if notification['state'] == 'paused':
                # Plex is a little weird and sometimes sends a session
                # on the WebSocket even though the HTTP API doesn't
                # return it. I've seen this happen after opening the
                # Plex web app and noticing that the mini-player at the
                # bottom of the page showed a paused episode. It's
                # probably getting reloaded from memory and sent as a
                # notification, but then they forget to update the API's
                # state. Anyway this is an icky situation and we'll get
                # notified if playback starts anyway, so just return
                # here.
                logger.debug(f"No session found for 'paused' notification")
                return
            elif notification['state'] == 'buffering':
                # Encountered this issue with an iPhone client that would buffer
                # a lot at the beginning of the session. To be investigated
                # further, but it might be that the HTTP API doesn't return a
                # session, until the first 'playing' notification. Not a huge
                # deal anyway as long as we get our 'playing' notification, so
                # let's just warn here.
                logger.warning(f"No session found for 'buffering' notification")
                return
            raise

        self._dispatch_and_schedule_extrapolated(session)



import argparse
from functools import partial
import logging
import os
from pathlib import Path
import shelve
import sys
import tempfile
from typing import Optional
import webbrowser

from pid import PidFile, PidFileError
from plexapi.myplex import MyPlexAccount, MyPlexResource
import pychromecast
import xdg
import zeroconf

from .auth import PlexApplication, PlexAuthClient
from .core import AutoSkipper
from .notifications import NotificationListener
from .seekables import (
    ChromecastMonitor,
    ChromecastSeekableProvider,
    PlexSeekableProvider,
    SeekableProviderChain
)
from .sessions import SessionDiscovery, SessionDispatcher, SessionProvider
from .stores import Database


# Note: Don't assume that the XDG paths are all different from each other (see
# Dockerfile).

if os.getenv('SK_DEV', '0') == '1':
    _DATABASE_PATH = xdg.xdg_data_home() / 'skippex_dev.db'
    _PID_NAME = 'skippex_dev.pid'
else:
    _DATABASE_PATH = xdg.xdg_data_home() / 'skippex.db'
    _PID_NAME = 'skippex.pid'

_APP_NAME = 'Skippex'
_APP_ARGV0 = __package__
_PID_DIR = xdg.xdg_runtime_dir() or Path(tempfile.gettempdir())
_PID_PATH = _PID_DIR / _PID_NAME


EXIT_UNAUTHORIZED = 4

logger = logging.getLogger(__name__)


def cmd_debug_info(args: argparse.Namespace, db: Database, app: PlexApplication):
    from pprint import pprint

    print(f'PID path: {_PID_PATH}')
    print(f'Database path: {_DATABASE_PATH}')
    print()
    print(f'Database content:')
    pprint(db.content())


def cmd_auth(args: argparse.Namespace, db: Database, app: PlexApplication):
    plex_auth = PlexAuthClient(app)
    pin_id, pin_code = plex_auth.generate_pin()
    auth_url = plex_auth.generate_auth_url(pin_code)

    webbrowser.open_new_tab(auth_url)
    logger.info('Navigate to the following page to authorize this application:')
    logger.info(auth_url)

    logger.info('Waiting for successful authorization...')
    auth_token = plex_auth.wait_for_token(pin_id, pin_code)

    db.auth_token = auth_token
    logger.info('Authorization successful')


def _find_server(account: MyPlexAccount, server_name: Optional[str]) -> Optional[MyPlexResource]:
    for resource in account.resources():
        if 'server' not in resource.provides:
            continue
        if (server_name and resource.name == server_name) or not server_name:
            return resource
    return None


def cmd_run(args: argparse.Namespace, db: Database, app: PlexApplication) -> Optional[int]:
    try:
        auth_token = db.auth_token
    except KeyError:
        logger.error(
            "No credentials found. Please first run the 'auth' command to "
            "authorize the application."
        )
        return EXIT_UNAUTHORIZED

    logger.info('Verifying token...')
    auth_client = PlexAuthClient(app)
    if not auth_client.is_token_valid(auth_token):
        logger.error("Token invalid. Please run the 'auth' command to reauthenticate yourself.")
        return EXIT_UNAUTHORIZED

    logger.info('Connecting to Plex server...')
    account = MyPlexAccount(token=auth_token)
    server_resource = _find_server(account, args.server)

    if not server_resource:
        if args.server:
            logger.error(f"Could not find server '{args.server}' for this account.")
        else:
            logger.error(f"Could not find a server associated with this account.")
        return 1

    # TODO: Ensure we try HTTP only if HTTPS fails.
    server = server_resource.connect()

    # Build the object hierarchy.

    cc_listener = pychromecast.discovery.CastListener()
    zconf = zeroconf.Zeroconf()
    cc_monitor = ChromecastMonitor(cc_listener, zconf)

    cc_listener.add_callback = cc_monitor.add_callback
    cc_listener.update_callback = cc_monitor.update_callback
    cc_listener.remove_callback = cc_monitor.remove_callback
    cc_browser = pychromecast.discovery.start_discovery(cc_listener, zconf)

    seekable_provider = SeekableProviderChain([
        PlexSeekableProvider(server),
        ChromecastSeekableProvider(cc_monitor),
    ])

    session_provider = SessionProvider(server)
    auto_skipper = AutoSkipper(seekable_provider)
    dispatcher = SessionDispatcher(listener=auto_skipper)

    discovery = SessionDiscovery(
        server=server,
        provider=session_provider,
        dispatcher=dispatcher,
        extrapolator=auto_skipper,
    )

    notif_listener = NotificationListener(server, discovery.alert_callback)
    logger.info('Ready')
    notif_listener.run_forever()


def _main():
    db = Database(shelve.open(str(_DATABASE_PATH)))
    app = PlexApplication(name=_APP_NAME, identifier=db.app_id)

    parser = argparse.ArgumentParser(_APP_ARGV0, formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--debug', help='enable debug logging', action='store_true')

    subparsers = parser.add_subparsers(title='subcommands', metavar='{auth,run}')
    subparsers.required = True

    parser_auth = subparsers.add_parser('auth', help='authorize this application to access your Plex account')
    parser_auth.set_defaults(func=partial(cmd_auth, db=db, app=app))

    parser_debug_info = subparsers.add_parser('debug-info')
    parser_debug_info.set_defaults(func=partial(cmd_debug_info, db=db, app=app))

    parser_run = subparsers.add_parser('run', help='monitor your shows and automatically skip intros')
    parser_run.set_defaults(func=partial(cmd_run, db=db, app=app))
    parser_run.add_argument('--server', help='name of your server (default: the first server Skippex finds)')

    args = parser.parse_args()

    if args.debug:
        log_level = logging.DEBUG
        log_format = '%(asctime)s - %(threadName)s - %(name)s - %(funcName)s - %(levelname)s - %(message)s'
        log_datefmt = None  # Use the default.
    else:
        log_level = logging.INFO
        log_format = '%(asctime)s - %(levelname)s - %(message)s'
        log_datefmt = '%Y-%m-%d %H:%M:%S'  # No milliseconds.

        # Disable logging from third-party packages.
        logger_name: str
        logger_inst: logging.Logger
        for logger_name, logger_inst in logging.root.manager.loggerDict.items():  # type: ignore
            if isinstance(logger_inst, logging.PlaceHolder):
                continue
            if not logger_name.startswith(__package__):
                logger_inst.addHandler(logging.NullHandler())
                logger_inst.propagate = False

    logging.basicConfig(
        level=log_level,
        format=log_format,
        datefmt=log_datefmt,
    )

    sys.exit(args.func(args))


def main():
    try:
        with PidFile(piddir=_PID_DIR, pidname=_PID_NAME):
            _main()
    except PidFileError:
        logger.error(
            f'Another instance of {_APP_NAME} is already running.\n'
            f'Please terminate it before running this command.'
        )
        return 1
    except KeyboardInterrupt:
        logger.info('Bye')
        return 0



from dataclasses import replace
import logging
from typing import Set, Tuple, cast

from .seekables import SeekableNotFoundError, SeekableProvider
from .sessions import (
    EpisodeSession,
    Session,
    SessionExtrapolator,
    SessionListener,
)


logger = logging.getLogger(__name__)


class AutoSkipper(SessionListener, SessionExtrapolator):
    def __init__(self, seekable_provider: SeekableProvider):
        self._skipped: Set[Session] = set()
        self._sp = seekable_provider

    def trigger_extrapolation(self, session: Session, listener_accepted: bool) -> bool:
        # Note it's only useful to do this when the state is 'playing':
        #  - When it's 'paused', we'll receive another notification either as
        #    soon as the state changes, or every 10 second while it's paused.
        #  - When it's 'buffering', we'll also receive another notification as
        #    soon as the state changes. And I assume we'd also get notified
        #    every 10 second otherwise.
        #  - When it's 'stopped', we've already sent a signal to the dispatcher.

        if not listener_accepted:
            logger.debug('No extrapolation: listener rejected')
            return False

        session = cast(EpisodeSession, session)  # Safe because listener_accepted.
        intro_marker = session.intro_marker()

        if session.view_offset_ms >= intro_marker.end:
            logger.debug('No extrapolation: beyond intro')
            return False

        # The listener accepted the session, and it may have skipped the intro.
        # In that case, we don't wanna extrapolate the session.
        return session not in self._skipped

    def extrapolate(self, session: Session) -> Tuple[Session, int]:
        session = cast(EpisodeSession, session)  # Safe thanks to trigger_extrapolation().
        delay_ms = 1000
        new_view_offset_ms = session.view_offset_ms + delay_ms
        return replace(session, view_offset_ms=new_view_offset_ms), delay_ms

    def accept_session(self, session: Session) -> bool:
        if not isinstance(session, EpisodeSession):
            # Only TV shows have intro markers, other media don't interest us.
            logger.debug('Ignored; not an episode')
            return False

        if session in self._skipped:
            logger.debug('Ignored; already skipped during this session')
            return False

        if session.state != 'playing':
            logger.debug(f'Ignored; state is "{session.state}" instead of "playing"')
            return False

        if not session.playable.hasIntroMarker:
            logger.debug(f'Ignored; has no intro marker')
            return False

        return True

    def on_session_activity(self, session: Session):
        session = cast(EpisodeSession, session)  # Safe thanks to accept_session().
        logger.debug(f'session_activity: {session}')

        intro_marker = session.intro_marker()
        view_offset_ms = session.view_offset_ms

        logger.debug(f'session.key={session.key}')
        logger.debug(f'session.view_offset_ms={session.view_offset_ms}')
        logger.debug(f'intro_marker={intro_marker}')

        if intro_marker.start <= view_offset_ms < intro_marker.end:
            try:
                seekable = self._sp.provide_seekable(session)
            except SeekableNotFoundError as e:
                if e.has_plex_player_not_found():
                    logger.error(
                        'Plex player not found for session; ensure "advertise '
                        'as player" is enabled'
                    )
                logger.exception(f'Cannot skip intro for session {session.key}')
                return

            seekable.seek(intro_marker.end)
            self._skipped.add(session)
            logger.info(f'Session {session.key}: skipped intro (seeked from {view_offset_ms} to {intro_marker.end})')
        else:
            logger.debug(f'Session {session.key}: did not skip (not viewing intro)')

        logger.debug('-----')

    def on_session_removal(self, session: Session):
        self._skipped.discard(session)



from time import sleep
from typing import Any, Dict, NamedTuple, Optional, Tuple
from urllib.parse import urlencode

import requests


class PlexApplication(NamedTuple):
    name: str
    identifier: str


class PlexAuthClient:
    # Reference: https://forums.plex.tv/t/authenticating-with-plex/609370

    _BASE_API_URL = 'https://plex.tv/api/v2'
    _BASE_AUTH_URL = 'https://app.plex.tv/auth'

    def __init__(self, app: PlexApplication):
        self._app = app

    def _make_request(self, method: str, endpoint: str, **kwargs: Dict[str, Any]) -> requests.Response:
        url = self._BASE_API_URL + endpoint
        return requests.request(method, url, **kwargs)

    def is_token_valid(self, token: str) -> bool:
        headers = {'Accept': 'application/json'}
        data = {
            'X-Plex-Product': self._app.name,
            'X-Plex-Client-Identifier': self._app.identifier,
            'X-Plex-Token': token,
        }
        r = self._make_request('GET', '/user', headers=headers, data=data)

        try:
            r.raise_for_status()
        except requests.HTTPError:
            if r.status_code == 401:
                return False
            raise
        else:
            return True

    def generate_pin(self) -> Tuple[int, str]:
        headers = {'Accept': 'application/json'}
        data = {
            'strong': 'true',
            'X-Plex-Product': self._app.name,
            'X-Plex-Client-Identifier': self._app.identifier,
        }
        r = self._make_request('POST', '/pins', headers=headers, data=data)
        r.raise_for_status()
        info = r.json()
        return info['id'], info['code']

    def generate_auth_url(self, pin_code: str) -> str:
        qs = urlencode({
            'clientID': self._app.identifier,
            'code': pin_code,
            'context[device][product]': self._app.name,
        })
        return self._BASE_AUTH_URL + '#?' + qs

    def check_pin(self, pin_id: int, pin_code: str) -> Optional[str]:
        headers = {'Accept': 'application/json'}
        data = {
            'code': pin_code,
            'X-Plex-Client-Identifier': self._app.identifier,
        }
        r = self._make_request('GET', f'/pins/{pin_id}', headers=headers, data=data)
        r.raise_for_status()
        info = r.json()
        return info['authToken'] if info['authToken'] else None

    def wait_for_token(self, pin_id: int, pin_code: str, check_interval_sec: int = 1) -> str:
        auth_token = self.check_pin(pin_id, pin_code)
        while not auth_token:
            sleep(check_interval_sec)
            auth_token = self.check_pin(pin_id, pin_code)
        return auth_token
